@@ -53,16 +53,20 @@ struct _EmpathyChatViewIface {
 	void             (*clear)                (EmpathyChatView *view);
 	gboolean         (*find_previous)        (EmpathyChatView *view,
 						  const gchar     *search_criteria,
-						  gboolean         new_search);
+						  gboolean         new_search,
+						  gboolean         match_case);
 	gboolean         (*find_next)            (EmpathyChatView *view,
 						  const gchar     *search_criteria,
-						  gboolean         new_search);
+						  gboolean         new_search,
+						  gboolean         match_case);
 	void             (*find_abilities)       (EmpathyChatView *view,
 						  const gchar     *search_criteria,
+						  gboolean         match_case,
 						  gboolean        *can_do_previous,
 						  gboolean        *can_do_next);
 	void             (*highlight)            (EmpathyChatView *view,
-						  const gchar     *text);
+						  const gchar     *text,
+						  gboolean         match_case);
 	void             (*copy_clipboard)       (EmpathyChatView *view);
 };
 
@@ -78,16 +82,20 @@ gboolean         empathy_chat_view_get_has_selection    (EmpathyChatView *view);
 void             empathy_chat_view_clear                (EmpathyChatView *view);
 gboolean         empathy_chat_view_find_previous        (EmpathyChatView *view,
 							 const gchar     *search_criteria,
-							 gboolean         new_search);
+							 gboolean         new_search,
+							 gboolean         match_case);
 gboolean         empathy_chat_view_find_next            (EmpathyChatView *view,
 							 const gchar     *search_criteria,
-							 gboolean         new_search);
+							 gboolean         new_search,
+							 gboolean         match_case);
 void             empathy_chat_view_find_abilities       (EmpathyChatView *view,
 							 const gchar     *search_criteria,
+							 gboolean         match_case,
 							 gboolean        *can_do_previous,
 							 gboolean        *can_do_next);
 void             empathy_chat_view_highlight            (EmpathyChatView *view,
-							 const gchar     *text);
+							 const gchar     *text,
+							 gboolean         match_case);
 void             empathy_chat_view_copy_clipboard       (EmpathyChatView *view);
 
 G_END_DECLS
