@@ -32,10 +32,12 @@
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
-#include <folks/folks.h>
 #include <telepathy-glib/account-manager.h>
 #include <telepathy-glib/util.h>
 
+#include <folks/folks.h>
+#include <folks/folks-telepathy.h>
+
 #include <libempathy/empathy-call-factory.h>
 #include <libempathy/empathy-individual-manager.h>
 #include <libempathy/empathy-contact-groups.h>
@@ -66,11 +68,11 @@ typedef struct
   EmpathyIndividualStore *store;
   GtkTreeRowReference *drag_row;
   EmpathyIndividualViewFeatureFlags view_features;
-  EmpathyContactFeatureFlags individual_features;
+  EmpathyIndividualFeatureFlags individual_features;
   GtkWidget *tooltip_widget;
-  GtkTargetList *file_targets;
 
   gboolean show_offline;
+  gboolean show_untrusted;
 
   GtkTreeModelFilter *filter;
   GtkWidget *search_widget;
@@ -101,6 +103,7 @@ enum
   PROP_VIEW_FEATURES,
   PROP_INDIVIDUAL_FEATURES,
   PROP_SHOW_OFFLINE,
+  PROP_SHOW_UNTRUSTED,
 };
 
 /* TODO: re-add DRAG_TYPE_CONTACT_ID, for the case that we're dragging around
@@ -108,6 +111,7 @@ enum
 enum DndDragType
 {
   DND_DRAG_TYPE_INDIVIDUAL_ID,
+  DND_DRAG_TYPE_PERSONA_ID,
   DND_DRAG_TYPE_URI_LIST,
   DND_DRAG_TYPE_STRING,
 };
@@ -116,20 +120,16 @@ enum DndDragType
   { (gchar *) T, 0, I }
 
 static const GtkTargetEntry drag_types_dest[] = {
+  DRAG_TYPE ("text/individual-id", DND_DRAG_TYPE_INDIVIDUAL_ID),
+  DRAG_TYPE ("text/persona-id", DND_DRAG_TYPE_PERSONA_ID),
   DRAG_TYPE ("text/path-list", DND_DRAG_TYPE_URI_LIST),
   DRAG_TYPE ("text/uri-list", DND_DRAG_TYPE_URI_LIST),
-  DRAG_TYPE ("text/contact-id", DND_DRAG_TYPE_INDIVIDUAL_ID),
   DRAG_TYPE ("text/plain", DND_DRAG_TYPE_STRING),
   DRAG_TYPE ("STRING", DND_DRAG_TYPE_STRING),
 };
 
-static const GtkTargetEntry drag_types_dest_file[] = {
-  DRAG_TYPE ("text/path-list", DND_DRAG_TYPE_URI_LIST),
-  DRAG_TYPE ("text/uri-list", DND_DRAG_TYPE_URI_LIST),
-};
-
 static const GtkTargetEntry drag_types_source[] = {
-  DRAG_TYPE ("text/contact-id", DND_DRAG_TYPE_INDIVIDUAL_ID),
+  DRAG_TYPE ("text/individual-id", DND_DRAG_TYPE_INDIVIDUAL_ID),
 };
 
 #undef DRAG_TYPE
@@ -139,7 +139,8 @@ static GdkAtom drag_atoms_source[G_N_ELEMENTS (drag_types_source)];
 
 enum
 {
-  DRAG_CONTACT_RECEIVED,
+  DRAG_INDIVIDUAL_RECEIVED,
+  DRAG_PERSONA_RECEIVED,
   LAST_SIGNAL
 };
 
@@ -176,7 +177,6 @@ individual_view_query_tooltip_cb (EmpathyIndividualView *view,
   GtkTreePath *path;
   static gint running = 0;
   gboolean ret = FALSE;
-  EmpathyContact *contact;
 
   priv = GET_PRIV (view);
 
@@ -203,17 +203,11 @@ individual_view_query_tooltip_cb (EmpathyIndividualView *view,
   if (individual == NULL)
     goto OUT;
 
-  contact = empathy_contact_dup_from_folks_individual (individual);
-  g_object_unref (individual);
-
-  if (contact == NULL)
-    goto OUT;
-
   if (priv->tooltip_widget == NULL)
     {
-      priv->tooltip_widget = empathy_contact_widget_new (contact,
-          EMPATHY_CONTACT_WIDGET_FOR_TOOLTIP |
-          EMPATHY_CONTACT_WIDGET_SHOW_LOCATION);
+      priv->tooltip_widget = empathy_individual_widget_new (individual,
+          EMPATHY_INDIVIDUAL_WIDGET_FOR_TOOLTIP |
+          EMPATHY_INDIVIDUAL_WIDGET_SHOW_LOCATION);
       gtk_container_set_border_width (GTK_CONTAINER (priv->tooltip_widget), 8);
       g_object_ref (priv->tooltip_widget);
       g_signal_connect (priv->tooltip_widget, "destroy",
@@ -221,12 +215,15 @@ individual_view_query_tooltip_cb (EmpathyIndividualView *view,
       gtk_widget_show (priv->tooltip_widget);
     }
   else
-    empathy_contact_widget_set_contact (priv->tooltip_widget, contact);
+    {
+      empathy_individual_widget_set_individual (
+        EMPATHY_INDIVIDUAL_WIDGET (priv->tooltip_widget), individual);
+    }
 
   gtk_tooltip_set_custom (tooltip, priv->tooltip_widget);
   ret = TRUE;
 
-  g_object_unref (contact);
+  g_object_unref (individual);
 OUT:
   running--;
 
@@ -249,44 +246,6 @@ groups_change_group_cb (GObject *source,
     }
 }
 
-static void
-individual_view_handle_drag (EmpathyIndividualView *self,
-    FolksIndividual *individual,
-    const gchar *old_group,
-    const gchar *new_group,
-    GdkDragAction action)
-{
-  g_return_if_fail (EMPATHY_IS_INDIVIDUAL_VIEW (self));
-  g_return_if_fail (FOLKS_IS_INDIVIDUAL (individual));
-
-  DEBUG ("individual %s dragged from '%s' to '%s'",
-      folks_individual_get_id (individual), old_group, new_group);
-
-  if (!tp_strdiff (new_group, EMPATHY_INDIVIDUAL_STORE_FAVORITE))
-    {
-      /* Mark contact as favourite */
-      folks_favourite_set_is_favourite (FOLKS_FAVOURITE (individual), TRUE);
-      return;
-    }
-
-  if (!tp_strdiff (old_group, EMPATHY_INDIVIDUAL_STORE_FAVORITE))
-    {
-      /* Remove contact as favourite */
-      folks_favourite_set_is_favourite (FOLKS_FAVOURITE (individual), FALSE);
-
-      /* Don't try to remove it */
-      old_group = NULL;
-    }
-
-  if (new_group != NULL)
-    folks_groups_change_group (FOLKS_GROUPS (individual), new_group, TRUE,
-        groups_change_group_cb, NULL);
-
-  if (old_group != NULL && action == GDK_ACTION_MOVE)
-    folks_groups_change_group (FOLKS_GROUPS (individual), old_group, FALSE,
-        groups_change_group_cb, NULL);
-}
-
 static gboolean
 group_can_be_modified (const gchar *name,
     gboolean is_fake_group,
@@ -309,20 +268,20 @@ group_can_be_modified (const gchar *name,
 }
 
 static gboolean
-individual_view_contact_drag_received (GtkWidget *self,
+individual_view_individual_drag_received (GtkWidget *self,
     GdkDragContext *context,
     GtkTreeModel *model,
     GtkTreePath *path,
     GtkSelectionData *selection)
 {
   EmpathyIndividualViewPriv *priv;
-  EmpathyIndividualManager *manager;
+  EmpathyIndividualManager *manager = NULL;
   FolksIndividual *individual;
   GtkTreePath *source_path;
   const gchar *sel_data;
   gchar *new_group = NULL;
   gchar *old_group = NULL;
-  gboolean new_group_is_fake, old_group_is_fake = TRUE;
+  gboolean new_group_is_fake, old_group_is_fake = TRUE, retval = FALSE;
 
   priv = GET_PRIV (self);
 
@@ -331,10 +290,16 @@ individual_view_contact_drag_received (GtkWidget *self,
       NULL, &new_group_is_fake);
 
   if (!group_can_be_modified (new_group, new_group_is_fake, TRUE))
-    return FALSE;
+    goto finished;
 
-  /* Get source group information. */
-  if (priv->drag_row)
+  /* Get source group information iff the view has the FEATURE_GROUPS_CHANGE
+   * feature. Otherwise, we just add the dropped contact to whichever group
+   * they were dropped in, and don't remove them from their old group. This
+   * allows for Individual views which shouldn't allow Individuals to have
+   * their groups changed, and also for dragging Individuals between Individual
+   * views. */
+  if ((priv->view_features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_GROUPS_CHANGE) &&
+      priv->drag_row != NULL)
     {
       source_path = gtk_tree_row_reference_get_path (priv->drag_row);
       if (source_path)
@@ -344,16 +309,19 @@ individual_view_contact_drag_received (GtkWidget *self,
               NULL, &old_group_is_fake);
           gtk_tree_path_free (source_path);
         }
+
+      if (!group_can_be_modified (old_group, old_group_is_fake, FALSE))
+        goto finished;
+
+      if (!tp_strdiff (old_group, new_group))
+        goto finished;
     }
-
-  if (!group_can_be_modified (old_group, old_group_is_fake, FALSE))
-    return FALSE;
-
-  if (!tp_strdiff (old_group, new_group))
+  else if (priv->drag_row != NULL)
     {
-      g_free (new_group);
-      g_free (old_group);
-      return FALSE;
+      /* We don't allow changing Individuals' groups, and this Individual was
+       * dragged from another group in *this* Individual view, so we disallow
+       * the drop. */
+      goto finished;
     }
 
   /* XXX: for contacts, we used to ensure the account, create the contact
@@ -366,23 +334,130 @@ individual_view_contact_drag_received (GtkWidget *self,
   if (individual == NULL)
     {
       DEBUG ("failed to find drag event individual with ID '%s'", sel_data);
-
-      g_object_unref (manager);
-
-      return FALSE;
+      goto finished;
     }
 
   /* FIXME: We should probably wait for the cb before calling
    * gtk_drag_finish */
 
-  individual_view_handle_drag (EMPATHY_INDIVIDUAL_VIEW (self), individual,
-      old_group, new_group, gdk_drag_context_get_selected_action (context));
+  /* Emit a signal notifying of the drag. We change the Individual's groups in
+   * the default signal handler. */
+  g_signal_emit (self, signals[DRAG_INDIVIDUAL_RECEIVED], 0,
+      gdk_drag_context_get_selected_action (context), individual, new_group,
+      old_group);
 
-  g_object_unref (G_OBJECT (manager));
+  retval = TRUE;
+
+finished:
+  tp_clear_object (&manager);
   g_free (old_group);
   g_free (new_group);
 
-  return TRUE;
+  return retval;
+}
+
+static void
+real_drag_individual_received_cb (EmpathyIndividualView *self,
+    GdkDragAction action,
+    FolksIndividual *individual,
+    const gchar *new_group,
+    const gchar *old_group)
+{
+  DEBUG ("individual %s dragged from '%s' to '%s'",
+      folks_individual_get_id (individual), old_group, new_group);
+
+  if (!tp_strdiff (new_group, EMPATHY_INDIVIDUAL_STORE_FAVORITE))
+    {
+      /* Mark contact as favourite */
+      folks_favourite_set_is_favourite (FOLKS_FAVOURITE (individual), TRUE);
+      return;
+    }
+
+  if (!tp_strdiff (old_group, EMPATHY_INDIVIDUAL_STORE_FAVORITE))
+    {
+      /* Remove contact as favourite */
+      folks_favourite_set_is_favourite (FOLKS_FAVOURITE (individual), FALSE);
+
+      /* Don't try to remove it */
+      old_group = NULL;
+    }
+
+  if (new_group != NULL)
+    {
+      folks_groups_change_group (FOLKS_GROUPS (individual), new_group, TRUE,
+          groups_change_group_cb, NULL);
+    }
+
+  if (old_group != NULL && action == GDK_ACTION_MOVE)
+    {
+      folks_groups_change_group (FOLKS_GROUPS (individual), old_group,
+          FALSE, groups_change_group_cb, NULL);
+    }
+}
+
+static gboolean
+individual_view_persona_drag_received (GtkWidget *self,
+    GdkDragContext *context,
+    GtkTreeModel *model,
+    GtkTreePath *path,
+    GtkSelectionData *selection)
+{
+  EmpathyIndividualViewPriv *priv;
+  EmpathyIndividualManager *manager = NULL;
+  FolksIndividual *individual = NULL;
+  FolksPersona *persona = NULL;
+  const gchar *persona_uid;
+  GList *individuals, *l;
+  gboolean retval = FALSE;
+
+  priv = GET_PRIV (self);
+
+  persona_uid = (const gchar *) gtk_selection_data_get_data (selection);
+
+  /* FIXME: This is slow, but the only way to find the Persona we're having
+   * dropped on us. */
+  manager = empathy_individual_manager_dup_singleton ();
+  individuals = empathy_individual_manager_get_members (manager);
+
+  for (l = individuals; l != NULL; l = l->next)
+    {
+      GList *personas, *p;
+
+      personas = folks_individual_get_personas (FOLKS_INDIVIDUAL (l->data));
+
+      for (p = personas; p != NULL; p = p->next)
+        {
+          if (!tp_strdiff (folks_persona_get_uid (FOLKS_PERSONA (p->data)),
+              persona_uid))
+            {
+              persona = g_object_ref (p->data);
+              individual = g_object_ref (l->data);
+              goto got_persona;
+            }
+        }
+    }
+
+got_persona:
+  g_list_free (individuals);
+
+  if (persona == NULL || individual == NULL)
+    {
+      DEBUG ("Failed to find drag event persona with UID '%s'", persona_uid);
+    }
+  else
+    {
+      /* Emit a signal notifying of the drag. We change the Individual's groups in
+       * the default signal handler. */
+      g_signal_emit (self, signals[DRAG_PERSONA_RECEIVED], 0,
+          gdk_drag_context_get_selected_action (context), persona, individual,
+          &retval);
+    }
+
+  tp_clear_object (&manager);
+  tp_clear_object (&persona);
+  tp_clear_object (&individual);
+
+  return retval;
 }
 
 static gboolean
@@ -438,13 +513,17 @@ individual_view_drag_data_received (GtkWidget *view,
     {
       success = FALSE;
     }
-  else if (info == DND_DRAG_TYPE_INDIVIDUAL_ID
-      || info == DND_DRAG_TYPE_STRING)
+  else if (info == DND_DRAG_TYPE_INDIVIDUAL_ID)
     {
-      success = individual_view_contact_drag_received (view,
+      success = individual_view_individual_drag_received (view,
           context, model, path, selection);
     }
-  else if (info == DND_DRAG_TYPE_URI_LIST)
+  else if (info == DND_DRAG_TYPE_PERSONA_ID)
+    {
+      success = individual_view_persona_drag_received (view, context, model,
+          path, selection);
+    }
+  else if (info == DND_DRAG_TYPE_URI_LIST || info == DND_DRAG_TYPE_STRING)
     {
       success = individual_view_file_drag_received (view,
           context, model, path, selection);
@@ -457,7 +536,12 @@ individual_view_drag_data_received (GtkWidget *view,
 static gboolean
 individual_view_drag_motion_cb (DragMotionData *data)
 {
-  gtk_tree_view_expand_row (GTK_TREE_VIEW (data->view), data->path, FALSE);
+  if (data->view != NULL)
+    {
+      gtk_tree_view_expand_row (GTK_TREE_VIEW (data->view), data->path, FALSE);
+      g_object_remove_weak_pointer (G_OBJECT (data->view),
+          (gpointer *) &data->view);
+    }
 
   data->timeout_id = 0;
 
@@ -508,17 +592,73 @@ individual_view_drag_motion (GtkWidget *widget,
       gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget), NULL, 0);
       return FALSE;
     }
-  target = gtk_drag_dest_find_target (widget, context, priv->file_targets);
+  target = gtk_drag_dest_find_target (widget, context, NULL);
   gtk_tree_model_get_iter (model, &iter, path);
 
-  if (target == GDK_NONE)
+  if (target == drag_atoms_dest[DND_DRAG_TYPE_URI_LIST] ||
+      target == drag_atoms_dest[DND_DRAG_TYPE_STRING])
     {
-      /* If target == GDK_NONE, then we don't have a target that can be
-         dropped on a contact.  This means a contact drag.  If we're
-         pointing to a group, highlight it.  Otherwise, if the contact
-         we're pointing to is in a group, highlight that.  Otherwise,
+      /* This is a file drag, and it can only be dropped on contacts,
+       * not groups.
+       * If we don't have FEATURE_FILE_DROP, disallow the drop completely,
+       * even if we have a valid target. */
+      FolksIndividual *individual = NULL;
+      EmpathyCapabilities caps = EMPATHY_CAPABILITIES_NONE;
+
+      if (priv->view_features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_FILE_DROP)
+        {
+          gtk_tree_model_get (model, &iter,
+              EMPATHY_INDIVIDUAL_STORE_COL_INDIVIDUAL, &individual,
+              -1);
+        }
+
+      if (individual != NULL)
+        {
+          EmpathyContact *contact = NULL;
+
+          contact = empathy_contact_dup_from_folks_individual (individual);
+          caps = empathy_contact_get_capabilities (contact);
+
+          tp_clear_object (&contact);
+        }
+
+      if (individual != NULL &&
+          folks_individual_is_online (individual) &&
+          (caps & EMPATHY_CAPABILITIES_FT))
+        {
+          gdk_drag_status (context, GDK_ACTION_COPY, time_);
+          gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget),
+              path, GTK_TREE_VIEW_DROP_INTO_OR_BEFORE);
+        }
+      else
+        {
+          gdk_drag_status (context, 0, time_);
+          gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget), NULL, 0);
+          retval = FALSE;
+        }
+
+      if (individual != NULL)
+        g_object_unref (individual);
+    }
+  else if ((target == drag_atoms_dest[DND_DRAG_TYPE_INDIVIDUAL_ID] &&
+      (priv->view_features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_GROUPS_CHANGE ||
+       priv->drag_row == NULL)) ||
+      (target == drag_atoms_dest[DND_DRAG_TYPE_PERSONA_ID] &&
+       priv->view_features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_PERSONA_DROP))
+    {
+      /* If target != GDK_NONE, then we have a contact (individual or persona)
+         drag.  If we're pointing to a group, highlight it.  Otherwise, if the
+         contact we're pointing to is in a group, highlight that.  Otherwise,
          set the drag position to before the first row for a drag into
          the "non-group" at the top.
+         If it's an Individual:
+           We only highlight things if the contact is from a different
+           Individual view, or if this Individual view has
+           FEATURE_GROUPS_CHANGE. This prevents highlighting in Individual views
+           which don't have FEATURE_GROUPS_CHANGE, but do have
+           FEATURE_INDIVIDUAL_DRAG and FEATURE_INDIVIDUAL_DROP.
+         If it's a Persona:
+           We only highlight things if we have FEATURE_PERSONA_DROP.
        */
       GtkTreeIter group_iter;
       gboolean is_group;
@@ -551,44 +691,6 @@ individual_view_drag_motion (GtkWidget *widget,
               group_path, GTK_TREE_VIEW_DROP_BEFORE);
         }
     }
-  else
-    {
-      /* This is a file drag, and it can only be dropped on contacts,
-         not groups.
-       */
-      FolksIndividual *individual;
-      EmpathyCapabilities caps = EMPATHY_CAPABILITIES_NONE;
-
-      gtk_tree_model_get (model, &iter,
-          EMPATHY_INDIVIDUAL_STORE_COL_INDIVIDUAL, &individual, -1);
-      if (individual != NULL)
-        {
-          EmpathyContact *contact = NULL;
-
-          contact = empathy_contact_dup_from_folks_individual (individual);
-          caps = empathy_contact_get_capabilities (contact);
-
-          tp_clear_object (&contact);
-        }
-
-      if (individual != NULL &&
-          folks_individual_is_online (individual) &&
-          (caps & EMPATHY_CAPABILITIES_FT))
-        {
-          gdk_drag_status (context, GDK_ACTION_COPY, time_);
-          gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget),
-              path, GTK_TREE_VIEW_DROP_INTO_OR_BEFORE);
-        }
-      else
-        {
-          gdk_drag_status (context, 0, time_);
-          gtk_tree_view_set_drag_dest_row (GTK_TREE_VIEW (widget), NULL, 0);
-          retval = FALSE;
-        }
-
-      if (individual != NULL)
-        g_object_unref (individual);
-    }
 
   if (!is_different && !cleanup)
     return retval;
@@ -611,6 +713,7 @@ individual_view_drag_motion (GtkWidget *widget,
       dm = g_new0 (DragMotionData, 1);
 
       dm->view = EMPATHY_INDIVIDUAL_VIEW (widget);
+      g_object_add_weak_pointer (G_OBJECT (widget), (gpointer *) &dm->view);
       dm->path = gtk_tree_path_copy (path);
 
       dm->timeout_id = g_timeout_add_seconds (1,
@@ -800,10 +903,12 @@ individual_view_row_activated (GtkTreeView *view,
   EmpathyIndividualViewPriv *priv = GET_PRIV (view);
   FolksIndividual *individual;
   EmpathyContact *contact = NULL;
+  FolksPresenceType best_presence = FOLKS_PRESENCE_TYPE_UNSET;
   GtkTreeModel *model;
   GtkTreeIter iter;
+  GList *personas, *l;
 
-  if (!(priv->individual_features & EMPATHY_CONTACT_FEATURE_CHAT))
+  if (!(priv->individual_features & EMPATHY_INDIVIDUAL_FEATURE_CHAT))
     return;
 
   model = gtk_tree_view_get_model (GTK_TREE_VIEW (view));
@@ -814,7 +919,31 @@ individual_view_row_activated (GtkTreeView *view,
   if (individual == NULL)
     return;
 
-  contact = empathy_contact_dup_from_folks_individual (individual);
+  /* Determine which Persona to chat to, by choosing the most available one. */
+  personas = folks_individual_get_personas (individual);
+  for (l = personas; l != NULL; l = l->next)
+    {
+      FolksPresenceType presence;
+
+      if (!TPF_IS_PERSONA (l->data))
+        continue;
+
+      /* Only choose the contact if it has a higher presence than our current
+       * best choice of contact. */
+      presence = folks_presence_get_presence_type (FOLKS_PRESENCE (l->data));
+      if (folks_presence_typecmp (presence, best_presence) > 0)
+        {
+          TpContact *tp_contact;
+
+          tp_clear_object (&contact);
+          tp_contact = tpf_persona_get_contact (TPF_PERSONA (l->data));
+          contact = empathy_contact_dup_from_tp_contact (tp_contact);
+          empathy_contact_set_persona (contact, FOLKS_PERSONA (l->data));
+
+          best_presence = presence;
+        }
+    }
+
   if (contact != NULL)
     {
       DEBUG ("Starting a chat");
@@ -832,6 +961,7 @@ individual_view_call_activated_cb (EmpathyCellRendererActivatable *cell,
     const gchar *path_string,
     EmpathyIndividualView *view)
 {
+  EmpathyIndividualViewPriv *priv = GET_PRIV (view);
   GtkWidget *menu;
   GtkTreeModel *model;
   GtkTreeIter iter;
@@ -840,6 +970,9 @@ individual_view_call_activated_cb (EmpathyCellRendererActivatable *cell,
   GtkMenuShell *shell;
   GtkWidget *item;
 
+  if (!(priv->view_features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_INDIVIDUAL_CALL))
+    return;
+
   model = gtk_tree_view_get_model (GTK_TREE_VIEW (view));
   if (!gtk_tree_model_get_iter_from_string (model, &iter, path_string))
     return;
@@ -1358,16 +1491,25 @@ individual_view_expand_idle_cb (EmpathyIndividualView *self)
   g_signal_handlers_block_by_func (self,
     individual_view_row_expand_or_collapse_cb, GINT_TO_POINTER (FALSE));
 
-  gtk_tree_model_foreach (GTK_TREE_MODEL (priv->filter),
-      (GtkTreeModelForeachFunc) expand_idle_foreach_cb, self);
+  /* The store/filter could've been removed while we were in the idle queue */
+  if (priv->filter != NULL)
+    {
+      gtk_tree_model_foreach (GTK_TREE_MODEL (priv->filter),
+          (GtkTreeModelForeachFunc) expand_idle_foreach_cb, self);
+    }
 
   g_signal_handlers_unblock_by_func (self,
       individual_view_row_expand_or_collapse_cb, GINT_TO_POINTER (FALSE));
   g_signal_handlers_unblock_by_func (self,
       individual_view_row_expand_or_collapse_cb, GINT_TO_POINTER (TRUE));
 
-  g_object_unref (self);
   priv->expand_groups_idle_handler = 0;
+  g_object_unref (self);
+
+  /* Empty the table of groups to expand/contract, since it may contain groups
+   * which no longer exist in the tree view. This can happen after going
+   * offline, for example. */
+  g_hash_table_remove_all (priv->expand_groups);
 
   return FALSE;
 }
@@ -1405,7 +1547,7 @@ individual_view_row_has_child_toggled_cb (GtkTreeModel *model,
    * a hash table, and expand or contract them as appropriate all at once in
    * an idle handler which iterates over all the group rows. */
   if (g_hash_table_lookup_extended (priv->expand_groups, name, NULL,
-      &will_expand) == FALSE &&
+      &will_expand) == FALSE ||
       GPOINTER_TO_INT (will_expand) != should_expand)
     {
       g_hash_table_insert (priv->expand_groups, g_strdup (name),
@@ -1472,7 +1614,9 @@ individual_view_store_row_deleted_cb (GtkTreeModel *model,
 
 static gboolean
 individual_view_is_visible_individual (EmpathyIndividualView *self,
-    FolksIndividual *individual)
+    FolksIndividual *individual,
+    gboolean is_online,
+    gboolean is_searching)
 {
   EmpathyIndividualViewPriv *priv = GET_PRIV (self);
   EmpathyLiveSearch *live = EMPATHY_LIVE_SEARCH (priv->search_widget);
@@ -1481,11 +1625,18 @@ individual_view_is_visible_individual (EmpathyIndividualView *self,
 
   /* We're only giving the visibility wrt filtering here, not things like
    * presence. */
-  if (live == NULL || gtk_widget_get_visible (GTK_WIDGET (live)) == FALSE)
-    return TRUE;
+  if (priv->show_untrusted == FALSE &&
+      folks_individual_get_trust_level (individual) == FOLKS_TRUST_LEVEL_NONE)
+    {
+      return FALSE;
+    }
+
+  if (is_searching == FALSE)
+    return (priv->show_offline || is_online);
 
   /* check alias name */
   str = folks_individual_get_alias (individual);
+
   if (empathy_live_search_match (live, str))
     return TRUE;
 
@@ -1497,7 +1648,10 @@ individual_view_is_visible_individual (EmpathyIndividualView *self,
       gchar *dup_str = NULL;
       gboolean visible;
 
-      str = folks_persona_get_uid (l->data);
+      if (!TPF_IS_PERSONA (l->data))
+        continue;
+
+      str = folks_persona_get_display_id (l->data);
       p = strstr (str, "@");
       if (p != NULL)
         str = dup_str = g_strndup (str, p - str);
@@ -1540,10 +1694,8 @@ individual_view_filter_visible_func (GtkTreeModel *model,
 
   if (individual != NULL)
     {
-      if (is_searching == TRUE)
-        visible = individual_view_is_visible_individual (self, individual);
-      else
-        visible = (priv->show_offline || is_online);
+      visible = individual_view_is_visible_individual (self, individual,
+          is_online, is_searching);
 
       g_object_unref (individual);
 
@@ -1576,12 +1728,12 @@ individual_view_filter_visible_func (GtkTreeModel *model,
       if (individual == NULL)
         continue;
 
-      visible = individual_view_is_visible_individual (self, individual);
+      visible = individual_view_is_visible_individual (self, individual,
+          is_online, is_searching);
       g_object_unref (individual);
 
       /* show group if it has at least one visible contact in it */
-      if ((is_searching && visible) ||
-          (!is_searching && (priv->show_offline || is_online)))
+      if (visible == TRUE)
         return TRUE;
     }
 
@@ -1592,28 +1744,10 @@ static void
 individual_view_constructed (GObject *object)
 {
   EmpathyIndividualView *view = EMPATHY_INDIVIDUAL_VIEW (object);
-  EmpathyIndividualViewPriv *priv = GET_PRIV (view);
   GtkCellRenderer *cell;
   GtkTreeViewColumn *col;
   guint i;
 
-  priv->filter = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (
-      GTK_TREE_MODEL (priv->store), NULL));
-  gtk_tree_model_filter_set_visible_func (priv->filter,
-      individual_view_filter_visible_func, view, NULL);
-
-  g_signal_connect (priv->filter, "row-has-child-toggled",
-      G_CALLBACK (individual_view_row_has_child_toggled_cb), view);
-  gtk_tree_view_set_model (GTK_TREE_VIEW (view),
-      GTK_TREE_MODEL (priv->filter));
-
-  tp_g_signal_connect_object (priv->store, "row-changed",
-      G_CALLBACK (individual_view_store_row_changed_cb), view, 0);
-  tp_g_signal_connect_object (priv->store, "row-inserted",
-      G_CALLBACK (individual_view_store_row_changed_cb), view, 0);
-  tp_g_signal_connect_object (priv->store, "row-deleted",
-      G_CALLBACK (individual_view_store_row_deleted_cb), view, 0);
-
   /* Setup view */
   g_object_set (view,
       "headers-visible", FALSE,
@@ -1737,10 +1871,10 @@ individual_view_set_view_features (EmpathyIndividualView *view,
      is enabled).
    */
   gtk_tree_view_set_reorderable (GTK_TREE_VIEW (view),
-      (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_CONTACT_DRAG));
+      (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_INDIVIDUAL_DRAG));
 
   /* Update DnD source/dest */
-  if (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_CONTACT_DRAG)
+  if (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_INDIVIDUAL_DRAG)
     {
       gtk_drag_source_set (GTK_WIDGET (view),
           GDK_BUTTON1_MASK,
@@ -1754,7 +1888,7 @@ individual_view_set_view_features (EmpathyIndividualView *view,
 
     }
 
-  if (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_CONTACT_DROP)
+  if (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_INDIVIDUAL_DROP)
     {
       gtk_drag_dest_set (GTK_WIDGET (view),
           GTK_DEST_DEFAULT_ALL,
@@ -1769,7 +1903,7 @@ individual_view_set_view_features (EmpathyIndividualView *view,
 
   /* Update has-tooltip */
   has_tooltip =
-      (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_CONTACT_TOOLTIP) != 0;
+      (features & EMPATHY_INDIVIDUAL_VIEW_FEATURE_INDIVIDUAL_TOOLTIP) != 0;
   gtk_widget_set_has_tooltip (GTK_WIDGET (view), has_tooltip);
 }
 
@@ -1782,7 +1916,6 @@ individual_view_dispose (GObject *object)
   tp_clear_object (&priv->store);
   tp_clear_object (&priv->filter);
   tp_clear_pointer (&priv->tooltip_widget, gtk_widget_destroy);
-  tp_clear_pointer (&priv->file_targets, gtk_target_list_unref);
 
   empathy_individual_view_set_live_search (view, NULL);
 
@@ -1794,6 +1927,8 @@ individual_view_finalize (GObject *object)
 {
   EmpathyIndividualViewPriv *priv = GET_PRIV (object);
 
+  if (priv->expand_groups_idle_handler != 0)
+    g_source_remove (priv->expand_groups_idle_handler);
   g_hash_table_destroy (priv->expand_groups);
 
   G_OBJECT_CLASS (empathy_individual_view_parent_class)->finalize (object);
@@ -1823,6 +1958,9 @@ individual_view_get_property (GObject *object,
     case PROP_SHOW_OFFLINE:
       g_value_set_boolean (value, priv->show_offline);
       break;
+    case PROP_SHOW_UNTRUSTED:
+      g_value_set_boolean (value, priv->show_untrusted);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
       break;
@@ -1841,7 +1979,7 @@ individual_view_set_property (GObject *object,
   switch (param_id)
     {
     case PROP_STORE:
-      priv->store = g_value_dup_object (value);
+      empathy_individual_view_set_store (view, g_value_get_object (value));
       break;
     case PROP_VIEW_FEATURES:
       individual_view_set_view_features (view, g_value_get_flags (value));
@@ -1853,6 +1991,10 @@ individual_view_set_property (GObject *object,
       empathy_individual_view_set_show_offline (view,
           g_value_get_boolean (value));
       break;
+    case PROP_SHOW_UNTRUSTED:
+      empathy_individual_view_set_show_untrusted (view,
+          g_value_get_boolean (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
       break;
@@ -1884,14 +2026,26 @@ empathy_individual_view_class_init (EmpathyIndividualViewClass *klass)
    * won't be called. */
   tree_view_class->row_activated = individual_view_row_activated;
 
-  signals[DRAG_CONTACT_RECEIVED] =
-      g_signal_new ("drag-contact-received",
+  klass->drag_individual_received = real_drag_individual_received_cb;
+
+  signals[DRAG_INDIVIDUAL_RECEIVED] =
+      g_signal_new ("drag-individual-received",
       G_OBJECT_CLASS_TYPE (klass),
       G_SIGNAL_RUN_LAST,
-      0,
+      G_STRUCT_OFFSET (EmpathyIndividualViewClass, drag_individual_received),
       NULL, NULL,
-      _empathy_gtk_marshal_VOID__OBJECT_STRING_STRING,
-      G_TYPE_NONE, 3, EMPATHY_TYPE_CONTACT, G_TYPE_STRING, G_TYPE_STRING);
+      _empathy_gtk_marshal_VOID__UINT_OBJECT_STRING_STRING,
+      G_TYPE_NONE, 4, G_TYPE_UINT, FOLKS_TYPE_INDIVIDUAL,
+      G_TYPE_STRING, G_TYPE_STRING);
+
+  signals[DRAG_PERSONA_RECEIVED] =
+      g_signal_new ("drag-persona-received",
+      G_OBJECT_CLASS_TYPE (klass),
+      G_SIGNAL_RUN_LAST,
+      G_STRUCT_OFFSET (EmpathyIndividualViewClass, drag_persona_received),
+      NULL, NULL,
+      _empathy_gtk_marshal_BOOLEAN__UINT_OBJECT_OBJECT,
+      G_TYPE_BOOLEAN, 3, G_TYPE_UINT, FOLKS_TYPE_PERSONA, FOLKS_TYPE_INDIVIDUAL);
 
   g_object_class_install_property (object_class,
       PROP_STORE,
@@ -1899,7 +2053,7 @@ empathy_individual_view_class_init (EmpathyIndividualViewClass *klass)
           "The store of the view",
           "The store of the view",
           EMPATHY_TYPE_INDIVIDUAL_STORE,
-          G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+          G_PARAM_READWRITE));
   g_object_class_install_property (object_class,
       PROP_VIEW_FEATURES,
       g_param_spec_flags ("view-features",
@@ -1910,16 +2064,23 @@ empathy_individual_view_class_init (EmpathyIndividualViewClass *klass)
   g_object_class_install_property (object_class,
       PROP_INDIVIDUAL_FEATURES,
       g_param_spec_flags ("individual-features",
-          "Features of the contact menu",
+          "Features of the individual menu",
           "Flags for all enabled features for the menu",
           EMPATHY_TYPE_INDIVIDUAL_FEATURE_FLAGS,
-          EMPATHY_CONTACT_FEATURE_NONE, G_PARAM_READWRITE));
+          EMPATHY_INDIVIDUAL_FEATURE_NONE, G_PARAM_READWRITE));
   g_object_class_install_property (object_class,
       PROP_SHOW_OFFLINE,
       g_param_spec_boolean ("show-offline",
           "Show Offline",
           "Whether contact list should display "
           "offline contacts", FALSE, G_PARAM_READWRITE));
+  g_object_class_install_property (object_class,
+      PROP_SHOW_UNTRUSTED,
+      g_param_spec_boolean ("show-untrusted",
+          "Show Untrusted Individuals",
+          "Whether the view should display untrusted individuals; "
+          "those who could not be who they say they are.",
+          TRUE, G_PARAM_READWRITE));
 
   g_type_class_add_private (object_class, sizeof (EmpathyIndividualViewPriv));
 }
@@ -1931,6 +2092,9 @@ empathy_individual_view_init (EmpathyIndividualView *view)
       EMPATHY_TYPE_INDIVIDUAL_VIEW, EmpathyIndividualViewPriv);
 
   view->priv = priv;
+
+  priv->show_untrusted = TRUE;
+
   /* Get saved group states. */
   empathy_contact_groups_get_all ();
 
@@ -1940,10 +2104,6 @@ empathy_individual_view_init (EmpathyIndividualView *view)
   gtk_tree_view_set_row_separator_func (GTK_TREE_VIEW (view),
       empathy_individual_store_row_separator_func, NULL, NULL);
 
-  /* Set up drag target lists. */
-  priv->file_targets = gtk_target_list_new (drag_types_dest_file,
-      G_N_ELEMENTS (drag_types_dest_file));
-
   /* Connect to tree view signals rather than override. */
   g_signal_connect (view, "button-press-event",
       G_CALLBACK (individual_view_button_press_event_cb), NULL);
@@ -2219,7 +2379,7 @@ empathy_individual_view_get_individual_menu (EmpathyIndividualView *view)
 
   /* Remove contact */
   if (priv->view_features &
-      EMPATHY_INDIVIDUAL_VIEW_FEATURE_CONTACT_REMOVE &&
+      EMPATHY_INDIVIDUAL_VIEW_FEATURE_INDIVIDUAL_REMOVE &&
       flags & EMPATHY_INDIVIDUAL_MANAGER_CAN_REMOVE)
     {
 
@@ -2336,3 +2496,90 @@ empathy_individual_view_set_show_offline (EmpathyIndividualView *self,
   g_object_notify (G_OBJECT (self), "show-offline");
   gtk_tree_model_filter_refilter (priv->filter);
 }
+
+gboolean
+empathy_individual_view_get_show_untrusted (EmpathyIndividualView *self)
+{
+  g_return_val_if_fail (EMPATHY_IS_INDIVIDUAL_VIEW (self), FALSE);
+
+  return GET_PRIV (self)->show_untrusted;
+}
+
+void
+empathy_individual_view_set_show_untrusted (EmpathyIndividualView *self,
+    gboolean show_untrusted)
+{
+  EmpathyIndividualViewPriv *priv;
+
+  g_return_if_fail (EMPATHY_IS_INDIVIDUAL_VIEW (self));
+
+  priv = GET_PRIV (self);
+
+  priv->show_untrusted = show_untrusted;
+
+  g_object_notify (G_OBJECT (self), "show-untrusted");
+  gtk_tree_model_filter_refilter (priv->filter);
+}
+
+EmpathyIndividualStore *
+empathy_individual_view_get_store (EmpathyIndividualView *self)
+{
+  g_return_val_if_fail (EMPATHY_IS_INDIVIDUAL_VIEW (self), NULL);
+
+  return GET_PRIV (self)->store;
+}
+
+void
+empathy_individual_view_set_store (EmpathyIndividualView *self,
+    EmpathyIndividualStore *store)
+{
+  EmpathyIndividualViewPriv *priv;
+
+  g_return_if_fail (EMPATHY_IS_INDIVIDUAL_VIEW (self));
+  g_return_if_fail (store == NULL || EMPATHY_IS_INDIVIDUAL_STORE (store));
+
+  priv = GET_PRIV (self);
+
+  /* Destroy the old filter and remove the old store */
+  if (priv->store != NULL)
+    {
+      g_signal_handlers_disconnect_by_func (priv->store,
+          individual_view_store_row_changed_cb, self);
+      g_signal_handlers_disconnect_by_func (priv->store,
+          individual_view_store_row_deleted_cb, self);
+
+      g_signal_handlers_disconnect_by_func (priv->filter,
+          individual_view_row_has_child_toggled_cb, self);
+
+      gtk_tree_view_set_model (GTK_TREE_VIEW (self), NULL);
+    }
+
+  tp_clear_object (&priv->filter);
+  tp_clear_object (&priv->store);
+
+  /* Set the new store */
+  priv->store = store;
+
+  if (store != NULL)
+    {
+      g_object_ref (store);
+
+      /* Create a new filter */
+      priv->filter = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (
+          GTK_TREE_MODEL (priv->store), NULL));
+      gtk_tree_model_filter_set_visible_func (priv->filter,
+          individual_view_filter_visible_func, self, NULL);
+
+      g_signal_connect (priv->filter, "row-has-child-toggled",
+          G_CALLBACK (individual_view_row_has_child_toggled_cb), self);
+      gtk_tree_view_set_model (GTK_TREE_VIEW (self),
+          GTK_TREE_MODEL (priv->filter));
+
+      tp_g_signal_connect_object (priv->store, "row-changed",
+          G_CALLBACK (individual_view_store_row_changed_cb), self, 0);
+      tp_g_signal_connect_object (priv->store, "row-inserted",
+          G_CALLBACK (individual_view_store_row_changed_cb), self, 0);
+      tp_g_signal_connect_object (priv->store, "row-deleted",
+          G_CALLBACK (individual_view_store_row_deleted_cb), self, 0);
+    }
+}
